@@ -93,7 +93,7 @@ void GeneralOptions::initOptions(ProgramOptions::OptionContext& root, ProgramOpt
 			"Set number of threads to use\n"
 			"      Default: %D / Valid: 0 <= %A < 64", "<n>")
 		("portfolio,p", storeTo(thread->portfolio), 
-			"Portfolio file to use for all but the master thread", "<file>")
+			"Portfolio file to use for threads", "<file>")
 		("force-gp", flag(thread->forceGP), 
 			"Force guiding path scheme even if portfolio is used")
 		("copy-problem", flag(thread->distribute.copyProblem), 
