@@ -518,7 +518,7 @@ void DefaultOutput::printThreadStats(const SolveStats** stats, std::size_t num)
 void DefaultOutput::printParallelStats(const SolveStats& stats, bool accu) const {
 	if (stats.parallel) {
 		const ParallelStats& ps = *stats.parallel;
-		uint64 solved = ps.gps-uint64(ps.terminated);
+		uint64 solved = ps.gps ? ps.gps-uint64(ps.terminated) : 0;
 		if (!accu) {
 			printKeyValue("CPU Time",  "%.3fs\n", ps.cpuTime);
 			printKeyValue("Models", "%"PRIu64"\n", stats.models);
@@ -858,7 +858,7 @@ void JsonOutput::printThreadStats(const SolveStats** stats, std::size_t num) {
 void JsonOutput::printParallelStats(const SolveStats& stats) {
 	if (stats.parallel) {
 		const ParallelStats& ps = *stats.parallel;
-		uint64 solved = ps.gps-uint64(ps.terminated);
+		uint64 solved = ps.gps ? ps.gps-uint64(ps.terminated) : 0;
 		startObject("Basic");
 		printKeyValue("CPU Time", ps.cpuTime);
 		printKeyValue("Models", stats.models);
