@@ -410,6 +410,7 @@ public:
 
 	// FOR TESTING
 	PrgAtomNode*  getAtom(Var atomId) const { return atoms_[atomId]; }
+	PrgBodyNode*  getBody(Var bodyId) const { return bodies_[bodyId]; }
 	Var           getEqAtom(Var a)    const { return getEqNode(atoms_, a);  }
 	const LitVec& getCompute()        const { return compute_; }
 private:
@@ -424,7 +425,6 @@ private:
 	typedef std::multimap<uint32, uint32>   BodyIndex; // hash -> bodies[offset]
 	typedef BodyIndex::iterator             IndexIter;
 	typedef std::pair<IndexIter, IndexIter> BodyRange;
-	typedef std::map<uint32, uint32>        EqNodes;
 	typedef std::pair<PrgBodyNode*, uint32> Body;
 	// ------------------------------------------------------------------------
 	// Program definition
@@ -440,6 +440,7 @@ private:
 	bool          eqWeights(const PrgBodyNode& b, WeightLitVec& w, bool& sorted) const;
 	bool          mergeEqAtoms(Var a, Var root);
 	Var           getEqBody(Var b) const { return getEqNode(bodies_, b);}
+	void          updateRule(PrgRule& r);
 	template <class C>
 	Var getEqNode(C& vec, Var id)  const {
 		if (!vec[id]->eq()) return id;
@@ -489,10 +490,10 @@ private:
 	} *minimize_;               // list of minimize-rules
 	struct Incremental  {
 		Incremental();
-		bool frozen(Var atom) const;
 		uint32  startAtom_;       // first atom of current iteration
 		uint32  startVar_;        // first var of the current iteration
 		uint32  startAux_;        // first aux atom of current iteration
+		uint32  startScc_;        // first free scc number
 		VarVec  freeze_;          // list of frozen atoms
 		VarVec  unfreeze_;        // list of atom that are unfreezed in this iteration
 	}* incData_;                // additional state to handle incrementally defined programs 
@@ -585,7 +586,9 @@ public:
 	void    setLiteral(Literal x)   { lit_ = x; }
 	void    clearLiteral(bool clVal){ lit_ = negLit(sentVar); if (clVal) value_ = value_free; }
 	bool    setValue(ValueRep v)    {
-		if (v == value_weak_true && ignore() && !frozen()) { v = value_true; }
+		if (v == value_weak_true && ignore() && !frozen()) { 
+			v = value_true; 
+		}
 		if (value_ == value_free || v == value_ || (value_ == value_weak_true && v == value_true)) {
 			value_ = v;
 			return true;
@@ -635,6 +638,17 @@ public:
 		}
 	}
 	//@}
+protected:
+	void assign(const PrgNode* other) {
+		lit_   = other->lit_;
+		root_  = other->root_;
+		value_ = other->value_;
+		scc_   = other->scc_;
+		eq_    = other->eq_;
+		seen_  = other->seen_;
+		ignore_= other->ignore_;
+		frozen_= other->frozen_;
+	}
 private:
 	PrgNode(const PrgNode&);
 	PrgNode& operator=(const PrgNode&);
@@ -932,6 +946,12 @@ public:
 	 */
 	typedef std::pair<bool, uint32> SimpRes;
 	SimpRes simplifyBodies(Var atomId, ProgramBuilder& prg, bool strong);
+	void    assign(const PrgAtomNode* a) {
+		PrgNode::assign(a);
+		posDep = a->posDep;
+		negDep = a->negDep;
+		preds  = a->preds;
+	}
 	VarVec    posDep; // Bodies in which this atom occurs positively
 	VarVec    negDep; // Bodies in which this atom occurs negatively
 	HeadVec   preds;  // Bodies having this atom as head
