@@ -555,14 +555,14 @@ public:
 private:
 	SharedContext(const SharedContext&);
 	SharedContext& operator=(const SharedContext&);
+	typedef std::auto_ptr<Distributor> DistPtr;
+	typedef std::auto_ptr<Enumerator>  EnumPtr;
 	struct SharedSymTab {
 		SharedSymTab() : refs(1) {}
 		SymbolTable symTab;
 		uint32      refs;
-	}* symTabPtr_;
-	typedef std::auto_ptr<Distributor> DistPtr;
-	typedef std::auto_ptr<Enumerator>  EnumPtr;
-	ProblemStats problem_;     // problem statistics
+	}*           symTabPtr_;   // pointer to shared symbol table
+	ProblemStats problem_;     // problem statistics1
 	VarInfo      varInfo_;     // varInfo[v] stores info about variable v
 	BTIG         btig_;        // binary-/ternary implication graph
 	Solver*      master_;      // master solver, responsible for init
