@@ -98,9 +98,9 @@ struct SolverStrategies {
 				}
 				// lbd scores are equal - fall through
 			case score_both: 
-			default: return ((int)lhs.scoreBoth()) - ((int)rhs.scoreBoth());
+			default: break;
 		}
-		return 0;
+		return ((int)lhs.scoreBoth()) - ((int)rhs.scoreBoth());
 	}
 	uint32              score(const LearntConstraint::Activity& act) const {
 		switch (reduceScore) {
@@ -400,13 +400,12 @@ public:
 		return force(p, r, d);
 	}
 
-	//! assumes the literal p.
+	//! assumes the literal p if possible
 	/*!
-	 * Sets the literal p to true and starts a new decision level.
-	 * \pre value(p) == value_free && validVar(p.var()) == true
+	 * If p is currently unassigned, sets p to true and starts a new decision level.
+	 * \pre validVar(p.var()) == true
 	 * \param p the literal to assume
-	 * \return always true
-	 * \post if decisionLevel() was n before the call then decisionLevel() is n + 1 afterwards.
+	 * \return !isFalse(p)
 	 */
 	bool assume(const Literal& p);
 
@@ -1048,8 +1047,8 @@ public:
 	 * Called whenever the solver must pick a new variable to branch on. 
 	 * \param s The solver that needs a new decision variable.
 	 * \return
-	 *  - true  : if the decision heuristic asserted a literal 
-	 *  - false : if no appropriate decision literal could be found, because assignment is total
+	 *  - true  : if the decision heuristic assumed a literal 
+	 *  - false : if no decision could be made because assignment is total or there is a conflict
 	 *  .
 	 * \post
 	 * if true is returned, the heuristic has asserted a literal.
