@@ -311,7 +311,7 @@ void ClaspFacade::solve(Input& problem, ClaspConfig& config, Callback* c) {
 	else if (!config.enumerate.onlyPre) {
 		setProblemSize();
 		AutoState state(this, state_solve);
-		ctrl_   = createSolveObject(config); // deleted by AutoState
+		initSolveObject(config); // deleted by AutoState
 		more_   = ctrl_->solve(config.ctx, master->solve, LitVec());
 	}
 }
@@ -336,7 +336,7 @@ void ClaspFacade::solveIncremental(Input& problem, ClaspConfig& config, Incremen
 			AutoState state(this, state_solve);
 			assume.clear();
 			problem.getAssumptions(assume);
-			ctrl_    = createSolveObject(config);
+			initSolveObject(config);
 			more_    = ctrl_->solve(config.ctx, master->solve, assume);
 			if (result_ == result_unknown && !more_) {
 				// initial assumptions are unsat
@@ -515,41 +515,49 @@ void ClaspFacade::setProblemSize() const {
 
 #ifndef DISABLE_MULTI_THREADING
 uint32 ClaspFacade::hardware_concurrency() { return mt::ParallelSolve::hardwareThreads(); }
-SolveAlgorithm* ClaspFacade::createSolveObject(ClaspConfig& config) const {
-	if (config.numThreads() == 1) {
-		return new SimpleSolve();
+void ClaspFacade::initSolveObject(ClaspConfig& config) {
+	ctrl_    = 0;
+	int numT = config.numThreads();
+	if (step_ != 0) {
+		assert(config.master()->solver().id() != mt::ParallelSolve::invalidId);
+		int invalid = 0;
+		for (uint32 i = 1; i != config.numThreads(); ++i) {
+			invalid += (config.threadConfig(i)->solver().id() == mt::ParallelSolve::invalidId);
+		}
+		numT -= invalid;
+	}
+	if (numT == 1) {
+		ctrl_ = new SimpleSolve();
+		return;
 	}
 	// parallel solving
-	int numT = config.numThreads();
-	std::auto_ptr<mt::ParallelSolve> ctrl(new mt::ParallelSolve());
-	ctrl->setForceGP(config.thread.forceGP);
-	ctrl->setIntegrate(config.thread.distribute.grace, config.thread.distribute.filter);
-	ctrl->setRestarts(config.thread.restarts.maxR, config.thread.restarts.sched);
-	SharedContext* ctx   = ctrl->init(numT, config.ctx);
-	LocalOptions* master = config.master();
-	bool enableJumpStats = master->solver().stats.jumps != 0;
-	for (int i = 1; i != numT; ++i) {
-		LocalOptions* ti = config.threadConfig(i);
-		if (enableJumpStats) {
-			ti->solver().stats.enableJumpStats();
-		}
-		if (graph_.get()) {
-			DefaultUnfoundedCheck* u = new DefaultUnfoundedCheck(ti->loopRep);
-			u->attachTo(ti->solver(), graph_.get());
-		}
-		ctrl->addThread(ti->solver(), ctx, ti->solve);
-	}
+	mt::ParallelSolve* ctrl = new mt::ParallelSolve();
+	ctrl_ = ctrl;
 	std::auto_ptr<mt::GlobalQueue> dist;
 	if (config.thread.distribute.types != 0) {
 		dist.reset(new mt::GlobalQueue(numT, config.thread.distribute.types, config.thread.distribute.lbd));
-		ctx->enableLearntSharing(dist.release());
+		config.ctx.enableLearntSharing(dist.release());
+	}
+	ctrl->setForceGP(config.thread.forceGP);
+	ctrl->setIntegrate(config.thread.distribute.grace, config.thread.distribute.filter);
+	ctrl->setRestarts(config.thread.restarts.maxR, config.thread.restarts.sched);
+	ctrl->init(numT, config.ctx);
+	for (int i = 1; i != numT; ++i) {
+		LocalOptions* ti = config.threadConfig(i);
+		ti->solver().stats.reset();
+		if (ti->solver().id() != mt::ParallelSolve::invalidId) {
+			if (ti->solver().sharedContext() == 0 && graph_.get()) {
+				DefaultUnfoundedCheck* u = new DefaultUnfoundedCheck(ti->loopRep);
+				u->attachTo(ti->solver(), graph_.get());
+			}
+			ctrl->addThread(ti->solver(), ti->solve);
+		}
 	}
-	return ctrl.release();
 }
 #else
 uint32 ClaspFacade::hardware_concurrency() { return 1; }
-SolveAlgorithm* ClaspFacade::createSolveObject(ClaspConfig&) const {
-	return new SimpleSolve();
+void ClaspFacade::initSolveObject(ClaspConfig&) {
+	ctrl_ = new SimpleSolve();
 }
 #endif
 
