@@ -608,7 +608,7 @@ bool Clause::minimize(Solver& s, Literal p, CCMinRecursive* rec) {
 			if (!s.ccMinimize(~*r, rec)) return false;
 		}
 		if (contracted()) {
-			for (const Literal* r = t.second; (ok = s.ccMinimize(~*r, rec)) && !r->watched(); ) {
+			for (const Literal* r = t.second; (ok = s.ccMinimize(~*r, rec)) == true && !r->watched(); ) {
 				++r;
 			}
 		}
