@@ -353,15 +353,14 @@ Literal UnitHeuristic::heuristic(Solver& s) {
 		} while (++i != sc.deps.size() && ok); 
 		if (!ok) {
 			// One of the candidates failed. Since none of them failed
-			// during previous propagation, this indicates that some
-			// post propagator has wrong priority. 
+			// during previous propagation, this indicates that 
+			// either some post propagator has wrong priority or
+			// parallel solving is active and a stop conflict was set.
 			// Since we can't resolve the problem here, we simply return the
-			// literal that caused the conflict in the hope that
-			// it is immediately regenerated once the literal is used as assumption.
+			// literal that caused the conflict
 			sc.clearDeps();
-			assert(s.hasConflict());
-			choice = s.decision(s.decisionLevel());
-			s.undoUntil(s.decisionLevel()-1);
+			assert(s.hasConflict() && s.decisionLevel() > 0);
+			choice = ~s.decision(s.decisionLevel());
 		}
 		sc.deps.clear();	
 	}
