@@ -22,122 +22,160 @@
 #include <clasp/solver.h>
 #include <clasp/clause.h>
 #include <clasp/clasp_facade.h>
-#include <clasp/util/thread.h>
 #include <clasp/util/timer.h>
 #include <clasp/minimize_constraint.h>
 #include <tbb/concurrent_queue.h>
 
+#if !defined(NDEBUG) && !defined(ENABLE_LOG)
+#define ENABLE_LOG 1
+#endif
+
+#if defined(ENABLE_LOG) && ENABLE_LOG == 1
+#include <stdio.h>
+#define LOG(id, fmt, ...) ( fprintf(stderr, "THREAD[%u]: " fmt "\n", (id), ##__VA_ARGS__), fflush(stderr) )
+#else
+#define LOG(id, fmt, ...)
+#endif
+
 namespace Clasp { namespace mt {
 /////////////////////////////////////////////////////////////////////////////////////////
-// Semaphore
+// BarrierSemaphore
 /////////////////////////////////////////////////////////////////////////////////////////
-class Semaphore {
+// A combination of a barrier and a semaphore
+class BarrierSemaphore {
 public:
-	explicit Semaphore(int counter = 0) : counter_(counter) {}
-	// Sets the internal counter to c.
+	explicit BarrierSemaphore(int counter = 0, int maxParties = 1) : counter_(counter), active_(maxParties) {}
+	// Initializes this object
 	// PRE: no thread is blocked on the semaphore
 	//      (i.e. internal counter is >= 0)
 	// NOTE: not thread-safe
-	void unsafe_reset(int c = 0) {
-		assert(counter_ >= 0);
-		counter_ = c;
+	void unsafe_init(int counter = 0, int maxParties = 1) {
+		counter_ = counter;
+		active_  = maxParties;
 	}
-
+	// Returns the current semaphore counter.
+	int  counter()   { std::unique_lock<tbb::mutex> lock(semMutex_); return counter_; }
+	// Returns the number of parties required to trip this barrier.
+	int  parties()   { std::unique_lock<tbb::mutex> lock(semMutex_); return active_;  } 
+	// Returns true if all parties are waiting at the barrier
+	bool active()    { std::unique_lock<tbb::mutex> lock(semMutex_); return unsafe_active(); }
+	
+	// barrier interface
+	
+	// Increases the barrier count, i.e. the number of 
+	// parties required to trip this barrier.
+	void addParty() {
+		std::unique_lock<tbb::mutex> lock(semMutex_);
+		++active_;
+	}
+	// Decreases the barrier count and resets the barrier
+	// if reset is true. 
+	// PRE: the thread does not itself wait on the barrier
+	void removeParty(bool reset) {
+		std::unique_lock<tbb::mutex> lock(semMutex_);
+		assert(active_ > 0);
+		--active_;
+		if      (reset)           { unsafe_reset(0); }
+		else if (unsafe_active()) { counter_ = -active_; lock.unlock(); semCond_.notify_one(); }
+	}
+	// Waits until all parties have arrived, i.e. called wait.
+	// Exactly one of the parties will receive a return value of true,
+	// the others will receive a value of false.
+	// Applications shall use this value to designate one thread as the
+	// leader that will eventually reset the barrier thereby unblocking the other threads.
+	bool wait() {
+		std::unique_lock<tbb::mutex> lock(semMutex_);
+		if (--counter_ >= 0) { counter_ = -1; }
+		return unsafe_wait(lock);
+	}
+	// Resets the barrier and unblocks any waiting threads.
+	void reset(uint32 semCount = 0) {
+		std::unique_lock<tbb::mutex> lock(semMutex_);
+		unsafe_reset(semCount);
+	}
+	// semaphore interface
+	
 	// Decrement the semaphore's counter.
 	// If the counter is zero or less prior to the call
 	// the calling thread is suspended.
-	void down() {
+	// Returns false to signal that all but the calling thread
+	// are currently blocked.
+	bool down() {
 		std::unique_lock<tbb::mutex> lock(semMutex_);
-		if (--counter_ < 0) semCond_.wait(lock);
+		if (--counter_ >= 0) { return true; }
+		return !unsafe_wait(lock);
 	}
-	// EFFECT: atomically
-	//  if ( (counter-1) > minC ) 
-	//    down(), return true;
-	//   return false;
-	bool try_down(int val = -1) {
-		std::unique_lock<tbb::mutex> lock(semMutex_);
-		if ((counter_-1) > val) {
-			if (--counter_ < 0) semCond_.wait(lock);
-			return true;
-		}
-		return false;
-	}
-	
-	// Increments the semaphore's counter
-	// and resumes one thread which has called down() 
-	// if the counter was less than zero prior to the call.
+	// Increments the semaphore's counter and resumes 
+	// one thread which has called down() if the counter 
+	// was less than zero prior to the call.
 	void up() {
 		bool notify;
 		{
 			std::unique_lock<tbb::mutex> lock(semMutex_);
-			notify = (++counter_ <= 0);
+			notify    = (++counter_ < 1);
 		}
 		if (notify) semCond_.notify_one();
 	}
-	
-	// Resumes currently blocked threads
-	// EFFECT: atomically executes
-	//   while (counter < 0) up();
-	void unblockAll() {
-		std::unique_lock<tbb::mutex> lock(semMutex_);
-		if (counter_ < 0) {
-			counter_ = 0;
-			semCond_.notify_all();
-		}
-	}
 private:
-	Semaphore(const Semaphore&);
-	Semaphore& operator=(const Semaphore&);
+	BarrierSemaphore(const BarrierSemaphore&);
+	BarrierSemaphore& operator=(const BarrierSemaphore&);
 	typedef std::condition_variable cv;
+	bool    unsafe_active() const { return -counter_ >= active_; }
+	void    unsafe_reset(uint32 semCount) {
+		int prev = counter_;
+		counter_ = semCount;
+		if (prev < 0) { semCond_.notify_all(); }
+	}
+	// Returns true for the leader, else false
+	bool    unsafe_wait(std::unique_lock<tbb::mutex>& lock) {
+		assert(counter_ < 0);
+		// don't put the last thread to sleep!
+		if (!unsafe_active()) {
+			semCond_.wait(lock);
+		}
+		return unsafe_active();
+	}	
 	cv         semCond_;  // waiting threads
 	tbb::mutex semMutex_; // mutex for updating counter
 	int        counter_;  // semaphore's counter
+	int        active_;   // number of active threads
 };
 /////////////////////////////////////////////////////////////////////////////////////////
 // ParallelSolve::Impl
 /////////////////////////////////////////////////////////////////////////////////////////
 struct ParallelSolve::SharedData {
 	typedef tbb::concurrent_queue<const LitVec*> queue;
-	SharedData() { reset(); }
-	void reset() {
-		workSem.unsafe_reset();
-		workQ.clear();
+	SharedData() : path(0) { reset(0, 1); }
+	void reset(SharedContext* a_ctx, uint32 numT) {
+		clearQueue();
+		init.reset();
+		workSem.unsafe_init(0, numT);
+		ctx    = a_ctx;
+		path   = 0;
+		nextId = 1;
 	}
-	tbb::mutex        modelM;      // model-mutex 
-	Semaphore         workSem;     // work-semaphore
-	queue             workQ;       // work-queue
-};
-
-/////////////////////////////////////////////////////////////////////////////////////////
-// Thread-state & message queue
-/////////////////////////////////////////////////////////////////////////////////////////
-// Thhread-local MessageQueue for storing some of the messages of Message_t
-struct MessageQueue {
-	enum MType { update_message = 0, message_types = 1 };
-	MessageQueue() { reset(); }
-	uint32 top()         const    { return time; }
-	uint32 top(MType m)  const    { return queue[m]; }
-	void   pop(uint32 t)          { time = t; }
-	void   pop(MType m, uint32 t) { queue[m] = t; }
-	void   reset()                {
-		restart = UINT64_MAX;
-		pop(0);
-		for (int i = 0; i != message_types; ++i) { pop(MType(i), 0); }
+	void clearQueue() {
+		for (const LitVec* a = 0; workQ.try_pop(a); ) {
+			if (a != path) { delete a; }
+		}
 	}
-	uint64 restart;
-	uint32 time;
-	uint32 queue[message_types];
+	Enumerator* enumerator() const { return ctx->enumerator(); }
+	SharedContext*    ctx;     // shared context object
+	const LitVec*     path;    // initial guiding path - typically empty
+	Timer<RealTime>   init;    // thread init time
+	tbb::mutex        modelM;  // model-mutex 
+	BarrierSemaphore  workSem; // work-semaphore
+	queue             workQ;   // work-queue
+	uint32            nextId;  // next solver id to use
 };
 
-struct ParallelSolve::ThreadInfo {
-	ThreadInfo() : thread(0), handler(0) {}
-	~ThreadInfo() { delete handler; delete thread; }
-	std::thread*     thread; // or 0 for master
-	ParallelHandler* handler;
-	MessageQueue     msg;
-	SolveStats       aggStats;
+// Callable function object to be used as "main" function
+// for threads. Simply forwards to ParallelSolve::doSolve()
+struct ParallelSolve::EntryPoint {
+	explicit EntryPoint(ParallelSolve* self) : self_(self) {}
+	void operator()(Solver* s, const SolveParams* p) { self_->doSolve(*s, *p, LitVec()); }
+	ParallelSolve* self_;
 };
-
 /////////////////////////////////////////////////////////////////////////////////////////
 // ParallelSolve
 /////////////////////////////////////////////////////////////////////////////////////////
@@ -169,9 +207,8 @@ ParallelSolve::ParallelSolve()
 	, shared_(new SharedData)
 	, thread_(0)
 	, globalR_(0,0.0,0)
+	, maxThreads_(0)
 	, maxRestarts_(0)
-	, numThreads_(1)
-	, nextId_(1)
 	, intGrace_(1024)
 	, intFlags_(ClauseCreator::clause_not_root_sat | ClauseCreator::clause_no_add)
 	, intHeuristic_(false)
@@ -179,32 +216,22 @@ ParallelSolve::ParallelSolve()
 }
 
 ParallelSolve::~ParallelSolve() {
+	if (maxThreads_ > 1) {
+		// algorithm was not started but there may be active threads -
+		// force orderly shutdown
+		ParallelSolve::terminate();
+		shared_->workSem.removeParty(true);
+		joinThreads();
+		assert(numThreads() == 0);
+	}
+	destroyThread(0);
 	delete shared_;
-	releaseThreads();
 }
 
 uint32 ParallelSolve::hardwareThreads() {
 	return std::thread::hardware_concurrency();
 }
 
-void ParallelSolve::releaseThreads() {
-	if (thread_) {
-		for (int i = 0; i != numThreads_; ++i) {
-			delete thread_[i];
-		}
-		delete [] thread_;
-		thread_ = 0;
-	}
-}
-
-void ParallelSolve::initThreads() {
-	releaseThreads();
-	thread_ = new ThreadInfo*[numThreads_];
-	for (int i = 0; i != numThreads_; ++i) {
-		thread_[i] = new ThreadInfo();
-	}
-}
-
 void ParallelSolve::setForceGP(bool gp) {
 	forceGP_ = gp;
 }
@@ -230,57 +257,115 @@ void ParallelSolve::setRestarts(uint32 maxR, const ScheduleStrategy& rs) {
 // prepare parallel solving
 SharedContext* ParallelSolve::init(int numT, SharedContext& ctx) {
 	assert(numT >= 1 && "Illegal number of threads");
-	shared_->reset();
-	exception_  = exception_none;
-	initialPath_= 0;
-	numThreads_ = static_cast<uint32>(numT);
+	shared_->reset(&ctx, static_cast<uint32>(numT));
 	msg_.reset();
-	initThreads();
+	reserveThreads();
+	shared_->init.start();
 	return &ctx;
 }
 
-// add new client thread
-void ParallelSolve::addThread(Solver& s, SharedContext* ctx, SolveParams& p) {
-	assert(thread_ && "ParallelSolve::init() not called!");
-	assert(nextId_ < ParallelSolve::numThreads() && (s.id() == 0 || s.id() == nextId_));
-	s.setId(nextId_++);
-	initThread(s, p);
-	thread_[s.id()]->thread  = new std::thread(std::bind2nd(std::mem_fun(&ParallelSolve::runThread), ThreadParams(ctx, s.id())), this);
+uint32 ParallelSolve::numThreads() const {
+	return shared_->workSem.parties();
 }
 
-// prepares s for solving by setting up its parallel handler
-void ParallelSolve::initThread(Solver& s, const SolveParams& p) {
-	assert(thread_ && "ParallelSolve::init() not called!");
-	ParallelHandler* h           = new ParallelHandler(this, s, p);
-	thread_[s.id()]->handler     = h;
-	if (s.stats.jumps)    thread_[s.id()]->aggStats.enableJumpStats();
-	s.stats.enableParallelStats();
-	thread_[s.id()]->aggStats.enableParallelStats();
-	thread_[s.id()]->msg.restart = globalR_.current();
-	if (s.id() == 0) {
-		h->init(s);
+void ParallelSolve::reserveThreads() {
+	destroyThread(0);
+	uint32 numT = numThreads();
+	thread_     = new ParallelHandler*[numT];
+	maxThreads_ = 0;
+	for (uint32 i = 0; i != numT; ++i) {
+#pragma message TODO("replace with CACHE_LINE_ALIGNED alloc")
+		uint32 b  = ((sizeof(ParallelHandler)+63) / 64) * 64;
+		thread_[i]= new (::operator new( b )) ParallelHandler(*this);
+		++maxThreads_;
 	}
 }
 
-void ParallelSolve::runThread(ThreadParams p) {
-	ParallelHandler* h = thread_[p.second]->handler;
-	SharedContext* ctx = p.first;
-	if (ctx->attach(h->solver())) {
-		doSolve(h->solver(), h->params(), LitVec());
-		ctx->detach(h->solver());
-	}
-	else {
-		terminateComplete(true);
-		endSolve(h->solver());
+void ParallelSolve::destroyThread(uint32 id) {
+	assert(id <= maxThreads_);
+	if (id < maxThreads_) {
+		assert(!thread_[id]->joinable() && "Shutdown not completed!");
+		thread_[id]->~ParallelHandler();
+		::operator delete(thread_[id]);
+		if (id == 0) {
+			assert(maxThreads_ == 1 && "Shutdown not completed!");
+			delete [] thread_;
+			thread_     = 0;
+			maxThreads_ = 0;
+		}
 	}
 }
 
+// joins with and destroys all active threads
+void ParallelSolve::joinThreads() {
+	shared_->init.reset();
+	shared_->init.start();
+	LOG(0, "STARTING  shutdown");
+	int error = thread_[0]->error();
+	for (uint32 i = 1, end = maxThreads_; i != end; ++i) {
+		if (thread_[i]->join() > error) {
+			error = thread_[i]->error();
+		}
+		destroyThread(i);
+	}
+	thread_[0]->setError(msg_.ok() ? thread_[0]->error() : error);
+	maxThreads_ = 1;
+	shared_->init.stop();
+	LOG(0, "COMPLETED shutdown after %.3fs", shared_->init.total());
+}
+
+// attach new client thread
+void ParallelSolve::addThread(Solver& s, SolveParams& p) {
+	uint32 id = shared_->nextId++;
+	assert(id < maxThreads_ && (s.id() == 0 || s.id() == id));
+	assert(thread_ && thread_[id]  && "ParallelSolve::init() not called!");
+	assert(!thread_[id]->joinable()&& "IDs are not unique or thread already initialized!");
+	s.setId(id);
+	thread_[id]->run( EntryPoint(this), s, p ); // start in a new thread
+}
+
+// called by all threads once before solving begins
+// establishes solver<->handler connection and attaches client
+// threads to shared context
+void ParallelSolve::startSolve(Solver& s, const SolveParams& p, const LitVec& assume) {
+	uint32 id = s.id();
+	LOG(id, "ENTERING  algorithm");
+	if (!msg_.terminate()) {
+		if (shared_->ctx->master()->stats.jumps) {
+			s.stats.enableJumpStats();
+		}
+		thread_[id]->attach(s, p); // connect solver<->handler
+		if (id == 0) {
+			// explicity init parallel handler because Solver::endInit() was
+			// already called.
+			thread_[0]->init(s);
+			shared_->path = &assume;
+		}
+		else if (!shared_->ctx->attach(s) && !msg_.quit()) {
+			// problem is UNSAT - terminate algorithm
+			LOG(s.id(), "TRIVIALLY UNSAT!");
+			terminateComplete(true);
+			return;
+		}
+		// wait for other threads
+		if (shared_->workSem.wait()) {
+			assert(msg_.workReq == 0 && shared_->workQ.empty());
+			initQueue(); // add initial path to work queue
+			shared_->init.stop();
+			LOG(id, "COMPLETED initialization after %.3fs", shared_->init.total());
+			shared_->workSem.reset();
+		}
+	}
+}
+
+// main solve loop executed by all threads
 bool ParallelSolve::doSolve(Solver& s, const SolveParams& p, const LitVec& assume) {
-	startSolve(s, p, assume);
-	MinimizeConstraint* min = 0;
+	PathPtr a(0);
 	try {
-		for (PathPtr a(0); requestWork(s, a);) {
-			thread_[s.id()]->aggStats.accu(s.stats);
+		startSolve(s, p, assume);
+		MinimizeConstraint* min = 0;
+		for (; requestWork(s, a);) {
+			thread_[s.id()]->prepareForGP(*a, globalR_.current(), a.is_owner());
 			s.stats.reset();
 			if (initPath(s, *a) && (!min || min->integrateNext(s))) {
 				if (solvePath(s, p) == value_free) {
@@ -289,73 +374,83 @@ bool ParallelSolve::doSolve(Solver& s, const SolveParams& p, const LitVec& assum
 				}
 				s.clearStopConflict();
 			}
-			if (min || enumerator()->optimizeHierarchical()) {
+			if (min || shared_->enumerator()->optimizeHierarchical()) {
 				// path is unsat and hierarchical optimization is active;
 				// relax minimize constraint to avoid problems during simplification
-				min  = enumerator()->constraint(s)->minimize();
+				min  = shared_->enumerator()->constraint(s)->minimize();
 				min->restoreOptimum();
 			}
 		}
 	}
-	catch (const std::bad_alloc&)     { terminate(); exception_ = exception_oom; }
-	catch (const std::runtime_error&) { terminate(); exception_ = exception_runtime; }
-	catch (const std::exception&)     { terminate(); exception_ = exception_other; }
+	catch (const std::bad_alloc&)      { exception(s,a,error_oom, "std::bad_alloc" ); }
+	catch (const std::runtime_error& e){ exception(s,a,error_runtime, e.what()); }
+	catch (...)                        { exception(s,a,error_other, "unknown");  }
 	return endSolve(s);
 }
 
-
-// called once by each solver before solving process starts
-// once all solvers are ready, we initialize the work-queue
-void ParallelSolve::startSolve(Solver& s, const SolveParams& p, const LitVec& assume) {
-	assert(thread_ && "ParallelSolve::init() not called!");
-	if (s.id() == 0) {
-		// attach the master thread
-		enumerator_  = s.sharedContext()->enumerator();
-		initialPath_ = &assume;
-		initThread(s, p);
-	}
-	// block on semaphore until all but one thread is waiting
-	if (!shared_->workSem.try_down(-numThreads_)) {
-		// all but this thread wait for work!
-		assert(msg_.workReq == 0 && shared_->workQ.empty());
-		initQueue();
-		// wake up all blocked threads
-		shared_->workSem.unblockAll();
-	}
-}
-
 // called once by each solver just after solving process has stopped
+// detaches the solver<->handler connection and joins with
+// all client threads
 bool ParallelSolve::endSolve(Solver& s) {
-	assert(msg_.terminate());
-	thread_[s.id()]->handler->release();
-	if (!msg_.quit()) {
+	assert(msg_.terminate() || thread_[s.id()]->error() != error_none);
+	LOG(s.id(), "LEAVING   algorithm");
+	uint32 id    = s.id();
+	int    error = thread_[id]->error();
+	thread_[id]->detach();
+	if (!msg_.quit() && s.sharedContext() && error == error_none) {
 		s.clearAssumptions();
 		if (s.hasConflict()) {
 			msg_.setFlag(ThreadControlFlags::complete_flag);
 		}
 	}
-	s.stats.parallel->cpuTime = ThreadTime::getTime();
-	s.stats.accu(thread_[s.id()]->aggStats);
-	// wake up any waiting threads
-	shared_->workSem.unblockAll();
 	// this thread is leaving
-	shared_->workSem.up();
-	if (s.id() == 0) {
-		// join with the other solvers
-		for (int i = 1; i != numThreads_; ++i) {
-			thread_[i]->thread->join();
+	shared_->workSem.removeParty(msg_.terminate());
+	s.stats.reset();
+	if (error != error_none) {
+		// solver terminated because of an exception
+		s.reset();
+		s.setId(ParallelSolve::invalidId);
+		std::swap(s.stats.jumps, thread_[id]->aggStats.jumps);
+		std::swap(s.stats.parallel, thread_[id]->aggStats.parallel);
+	}
+	// copy aggregated statistics for this solver to solver
+	s.stats.accu(thread_[id]->aggStats);
+	id == 0
+		? joinThreads()            // join with the other solvers
+		: shared_->ctx->detach(s); // detach client threads from shared context object
+	if (s.stats.parallel) {
+		s.stats.parallel->cpuTime = ThreadTime::getTime();
+	}
+	if (id == 0) {
+		error = thread_[0]->error();
+		if (error != error_none) {
+			switch (error) {
+				case error_oom    : throw std::bad_alloc();
+				case error_runtime: throw std::runtime_error("RUNTIME ERROR!");
+				default:            throw std::runtime_error("UNKNOWN ERROR!");
+			}
 		}
-		switch (exception_) {
-			case exception_none   : break;
-			case exception_oom    : throw std::bad_alloc();
-			case exception_runtime: throw std::runtime_error("RUNTIME ERROR!");
-			default:                throw std::runtime_error("UNKNOWN ERROR!");
-		}
-		enumerator()->reportResult(msg_.complete());
+		shared_->enumerator()->reportResult(msg_.complete());
 	}
 	return !msg_.complete();
 }
 
+void ParallelSolve::exception(Solver& s, PathPtr& path, ErrorCode e, const char* what) {
+	try {
+		LOG(s.id(), "EXCEPTION %s", what); (void) what;
+		thread_[s.id()]->setError(e);
+		if (s.id() == 0 || shared_->workSem.active()) { 
+			terminate();
+			return;
+		}
+		else if (msg_.allowSplit() && path.get()) {
+			shared_->workQ.push(path.release());
+			shared_->workSem.up();
+		}
+	}
+	catch(...) { terminate(); }
+}
+
 // forced termination from outside
 bool ParallelSolve::terminate() {
 	// do not use postMessage() to avoid possible
@@ -376,11 +471,12 @@ void ParallelSolve::terminateComplete(bool complete) {
 // terminated because of unsat, check if there is more to do
 void ParallelSolve::terminateUnsat(Solver& s, bool complete) {
 	if (!msg_.terminate()) {
-		if (complete && enumerator()->optimizeHierarchical() && s.popRootLevel(s.rootLevel(), false)) {
+		if (complete && shared_->enumerator()->optimizeHierarchical() && s.popRootLevel(s.rootLevel(), false)) {
 			// Problem is unsat and hierarchical optimization is active.
 			// The active level is at its optimum, but lower-priority levels might
 			// still be non-optimal.
 			// Notify other threads to prepare for solving next priority level.
+			LOG(s.id(), "COMPLETED current optimization step"); 
 			postMessage(Message_t::unsat_message);
 			assert(s.decisionLevel() == 0);
 			return;
@@ -397,7 +493,7 @@ void ParallelSolve::postMessage(MessageType m) {
 		ThreadControlFlags::Flag cf = static_cast<ThreadControlFlags::Flag>((uint32(1)<<m));
 		if (msg_.setFlag(cf)) {
 			++msg_.messages;
-			shared_->workSem.unblockAll();
+			shared_->workSem.reset();
 		}
 	}
 	else if (m == Message_t::split_message) {
@@ -413,6 +509,7 @@ void ParallelSolve::postMessage(MessageType m) {
 	}
 }
 
+
 // tries to get new work for the given solver
 bool ParallelSolve::requestWork(Solver& s, PathPtr& out) { 
 	const LitVec* a;
@@ -427,13 +524,11 @@ bool ParallelSolve::requestWork(Solver& s, PathPtr& out) {
 		else if (shared_->workQ.try_pop(a)) {
 			assert(s.decisionLevel() == 0);
 			// got new work from work-queue
-			bool splitted = a != initialPath_;
-			out           = a;
-			if (!splitted) out.release(); // do not take over ownership of initial gp!
-			thread_[s.id()]->msg.restart = globalR_.current();
-			thread_[s.id()]->handler->prepareForGP(*a, splitted);
+			out = a;
+			// do not take over ownership of initial gp!
+			if (a == shared_->path) { out.release(); } 
 			// propagate any new facts now that we have new work
-			if (s.simplify()) { return true; }
+			if (s.simplify())      { return true; }
 			// s now has a conflict - either an artifical stop conflict
 			// or a real conflict - we'll handle it in the next iteration
 			// via the call to popRootLevel()
@@ -447,7 +542,7 @@ bool ParallelSolve::requestWork(Solver& s, PathPtr& out) {
 			// gp mode is active - request a split	
 			// and wait until someone has work for us
 			postMessage(Message_t::split_message);
-			if (!shared_->workSem.try_down(-numThreads_)) {
+			if (!shared_->workSem.down() && !msg_.restart()) {
 				// we are the last man standing, there is no
 				// work left - quitting time?
 				terminateUnsat(s, true);
@@ -467,41 +562,37 @@ bool ParallelSolve::requestWork(Solver& s, PathPtr& out) {
 bool ParallelSolve::waitOnRestart(Solver& s) {
 	assert(msg_.restart());
 	// react once
-	if (!thread_[s.id()]->handler->handleRestartMessage(s)) {
+	if (!thread_[s.id()]->handleRestartMessage()) {
 		msg_.setFlag(ThreadControlFlags::inhibit_restart_flag);
 	}
-	// loop because the semaphore counter may be positive
-	while (shared_->workSem.try_down(-numThreads_)) {
-		if (msg_.terminate()) {
-			return true;
+	LOG(s.id(), "WAITING   for restart");
+	if (shared_->workSem.wait()) {
+		// last man standing - complete restart request
+		msg_.workReq   = 0;
+		msg_.restartReq= 0;
+		bool hasUnsat  = msg_.unsatRestart();
+		msg_.control  -= (msg_.control & (uint32(ThreadControlFlags::unsat_flag|ThreadControlFlags::restart_flag)));
+		if (!hasUnsat) {
+			globalR_.next();
+			if ( (maxRestarts_ -= maxRestarts_ > 0) == 0 ) {
+				globalR_.init(0, 0.0, 0);
+			}
 		}
-		if (!msg_.restart()) {
-			return msg_.allowRestart();
+		if (hasUnsat || msg_.allowRestart()) {
+			initQueue();
 		}
-		// request still active - continue waiting
-	}
-	// last man standing - complete restart request
-	msg_.workReq   = 0;
-	msg_.restartReq= 0;
-	bool hasUnsat  = msg_.unsatRestart();
-	msg_.control  -= (msg_.control & (uint32(ThreadControlFlags::unsat_flag|ThreadControlFlags::restart_flag)));
-	if (!hasUnsat) {
-		globalR_.next();
-		if ( (maxRestarts_ -= maxRestarts_ > 0) == 0 ) {
-			globalR_.init(0, 0.0, 0);
+		if (hasUnsat && shared_->enumerator()->optimizeHierarchical()) {
+			Enumerator*         en = shared_->enumerator();
+			MinimizeConstraint* m  = en->constraint(s)->minimize();
+			if (en->enumerated == 0 || !m->shared_unsafe()->optimizeNext()) {
+				terminateComplete(true);
+			}
 		}
+		LOG(s.id(), "COMPLETED restart");
+		// wake up all blocked threads
+		shared_->workSem.reset();
 	}
-	if (hasUnsat || msg_.allowRestart()) {
-		initQueue();
-	}
-	if (hasUnsat && enumerator()->optimizeHierarchical()) {
-		if (enumerator()->enumerated == 0 || !enumerator()->constraint(s)->minimize()->shared_unsafe()->optimizeNext()) {
-			terminateComplete(true);
-		}
-	}
-	// wake up all blocked threads
-	shared_->workSem.unblockAll();
-	return msg_.terminate() || msg_.allowRestart();
+	return msg_.terminate() || (!msg_.restart() && msg_.allowRestart());
 }
 
 // If guiding path scheme is active only one
@@ -511,8 +602,8 @@ bool ParallelSolve::waitOnRestart(Solver& s) {
 // TODO:
 //  heuristic for initial splits?
 void ParallelSolve::initQueue() {
-	shared_->workQ.clear();
-	int end = numThreads_;
+	shared_->clearQueue();
+	int end = numThreads();
 	if (forceGP_) {
 		msg_.setFlag(ThreadControlFlags::allow_gp_flag);
 		end = 1;
@@ -521,7 +612,7 @@ void ParallelSolve::initQueue() {
 		msg_.setFlag(ThreadControlFlags::inhibit_restart_flag);
 	}
 	for (int i = 0; i != end; ++i) {
-		shared_->workQ.push(initialPath_);
+		shared_->workQ.push(shared_->path);
 	}
 }
 
@@ -544,26 +635,27 @@ bool ParallelSolve::backtrackFromModel(Solver& s) {
 		// first check if the model is still valid once all
 		// information is integrated into the solver
 		uint32 dl = s.decisionLevel();
-		if (!thread_[s.id()]->handler->isModel(s)) {
+		if (!thread_[s.id()]->isModel(s)) {
 			// model no longer a (unique) model - continue search
 			s.stats.removeModel(dl);
 			return true;
 		}
-		r = enumerator()->backtrackFromModel(s);
+		r = shared_->enumerator()->backtrackFromModel(s);
 		if (r == Enumerator::enumerate_stop_enough || (r == Enumerator::enumerate_stop_complete && s.decisionLevel() == 0)) {
 			// must be called while holding the lock - otherwise
 			// we have a race condition with solvers that
 			// are currently blocking on the mutex and we could enumerate 
-			// more model than requested by the user
+			// more models than requested by the user
 			terminateComplete(s.decisionLevel() == 0);
 		}
-		else if (!enumerator()->trivial(thread_[s.id()]->handler->disjointPath())) {
+		else if (!shared_->enumerator()->trivial(thread_[s.id()]->disjointPath())) {
 			// enumerator is not trivial w.r.t current search scheme
 			// force update of other solvers
 			postMessage(Message_t::update_message);
-			thread_[s.id()]->msg.pop(MessageQueue::update_message, msg_.updates);
+			// ignore message in the active solver
+			thread_[s.id()]->setUpdate(msg_.updates);
 		}
-		if (enumerator()->enumerated == 1 && !enumerator()->supportsRestarts()) {
+		if (shared_->enumerator()->enumerated == 1 && !shared_->enumerator()->supportsRestarts()) {
 			msg_.setFlag(ThreadControlFlags::inhibit_restart_flag);
 			msg_.setFlag(ThreadControlFlags::allow_gp_flag);
 		}
@@ -571,130 +663,161 @@ bool ParallelSolve::backtrackFromModel(Solver& s) {
 	return r == Enumerator::enumerate_continue && !msg_.terminate();
 }
 
+// checks for new update messages
+// POST: m.doUpdate if there is an active update message for m
+// return: m.doUpdate
+bool ParallelSolve::checkUpdates(Message_t::Local& m) const {
+	uint32 updates     = msg_.updates;
+	uint32 seen        = m.update;
+	m.update           = updates;
+	return (m.doUpdate = (m.doUpdate || seen != updates));
+}
+
 // updates s with new messages and uses s to create a new guiding path
 // if necessary and possible
-bool ParallelSolve::handleMessages(Solver& s) {
+bool ParallelSolve::handleMessages(Solver& s, Message_t::Local& m) {
 	// check if there are new messages for s
-	uint32 hId      = s.id();
-	uint32 gt       = msg_.messages;
-	MessageQueue& m = thread_[hId]->msg;
-	if (m.top() != gt) {
-		ParallelHandler* h = thread_[hId]->handler;
+	uint32 hId         = s.id();
+	uint32 gt          = msg_.messages;
+	ParallelHandler* h = thread_[hId];
+	if (m.any != gt) {
 		if (msg_.hasControlMsg()) {
 			if (msg_.terminate()) {
-				return h->handleTerminateMessage(s);
+				LOG(s.id(), "TERMINATE message received!");
+				h->handleTerminateMessage();
+				s.stats.parallel->terminated = true;
+				s.setStopConflict();
+				return false;
 			}
 			if (msg_.restart() && waitOnRestart(s)) {
 				s.setStopConflict();
 				return false;
 			}
+			return true;
 		}
-		uint32 mt;
-		if (m.top(MessageQueue::update_message) != (mt = msg_.updates) ) {
-			m.pop(MessageQueue::update_message, mt);
-			if (!h->handleModelMessage(s)) {
-				return false;
-			}
-		}
-		mt = msg_.workReq; 
+		checkUpdates(m);
+		uint32 mt  = msg_.workReq; 
 		if (mt <= 0) {
-			m.pop(gt);
+			m.any    = gt;
 		}
-		else if (s.splittable() && h->splittable()) {
+		else if (h->splittable()) {
 			// First declare split request as handled
 			// and only then do the actual split.
 			// This way, we minimize the chance for 
 			// "over"-splitting, i.e. one split request handled
 			// by more than one thread.
 			if (--msg_.workReq == 0) {
-				m.pop(gt);
+				m.any  = gt;
 			}
-			h->handleSplitMessage(s);
+			h->handleSplitMessage();
 		}
 	}
-	if (s.stats.conflicts >= m.restart) {
-		requestRestart();
-		m.restart *= 2;
-	}
 	return true;
 }
 
-bool ParallelSolve::validModel(Solver& s, bool forceUpdate, bool disjointPath) {
-	assert(s.numFreeVars() == 0);
-	uint32 hId      = s.id();
-	MessageQueue& m = thread_[hId]->msg;
-	uint32       mt = msg_.updates;
-	if (m.top(MessageQueue::update_message) != mt) {
-		m.pop(MessageQueue::update_message, mt);
-		forceUpdate = true;
-	}
-	if (forceUpdate && !enumerator()->update(s, disjointPath)) {
-		return false;
-	}
-	return s.numFreeVars() == 0 && !msg_.terminate();
-}
-
 void ParallelSolve::requestRestart() {
-	if (msg_.allowRestart() && ++msg_.restartReq == numThreads_) {
+	if (msg_.allowRestart() && ++msg_.restartReq == numThreads()) {
 		postMessage(Message_t::restart_message);
 	}
 }
-/////////////////////////////////////////////////////////////////////////////////////////
+
+////////////////////////////////////////////////////////////////////////////////////
 // ParallelHandler
 /////////////////////////////////////////////////////////////////////////////////////////
-ParallelHandler::ParallelHandler(ParallelSolve* ctrl, Solver& s, const SolveParams& p) 
-	: solver_(&s)
-	, params_(&p)
+ParallelHandler::ParallelHandler(ParallelSolve& ctrl) 
+	: solver_(0)
+	, params_(0)
 	, intTail_(0)
-	, messageHandler_(ctrl)
-	, updateModel_(false) {
+	, error_(0)
+	, messageHandler_(&ctrl) {
+}
+
+ParallelHandler::~ParallelHandler() {
+	detach();
+}
+
+// adds this as post propagator to s
+void ParallelHandler::attach(Solver& s, const SolveParams& p) {
+	detach();
+	aggStats.reset();
+	s.stats.enableParallelStats();
+	aggStats.enableParallelStats();
+	if (s.stats.jumps) { aggStats.enableJumpStats(); }
+	messageHandler_.msg = MessageHandler::MessageQ();
+	gp_.reset();
+	intTail_= 0;
+	error_  = 0;
 	s.addPost(&messageHandler_);
 	s.addPost(this);
+	solver_ = &s;
+	params_ = &p;
 }
 
-void ParallelHandler::prepareForGP(const LitVec& out, bool fromSplit) {
-	gp_.reset();
-	gp_.split = fromSplit;
-	solver_->stats.parallel->newGP(out.size());
+// detaches this from its solver
+void ParallelHandler::detach() {
+	if (solver_) {
+		handleTerminateMessage();
+		aggStats.accu(solver_->stats);
+		while (!integrated_.empty()) {
+			ClauseHead* c = (ClauseHead*)integrated_.back();
+			integrated_.pop_back();
+			if (error() == 0) {
+				if (c->locked(*solver_)) {
+					solver_->addLearnt(c, c->size());
+				}
+				else {
+					c->destroy(solver_, true);
+				}
+			}
+			else { c->destroy(0, false); }
+		}
+		integrated_.clear();
+		intTail_= 0;
+		solver_ = 0;
+		params_ = 0;
+	}
 }
 
-void ParallelHandler::detach(bool onyMessageHandler) {
-	// mark removed propagators by creating "self-loop"
-	if (messageHandler_.next != &messageHandler_) {
+// true if solver can be used to split-off a new guiding path
+bool ParallelHandler::splittable() const {
+	return solver_->splittable() && !messageHandler_.pendingUpdate();
+}
+
+void ParallelHandler::prepareForGP(const LitVec& out, uint64 restart, bool fromSplit) {
+	gp_.reset(restart, fromSplit);
+	aggStats.parallel->newGP(out.size());
+	aggStats.accu(solver_->stats);
+}
+
+// detach from solver, i.e. ignore any further messages 
+void ParallelHandler::handleTerminateMessage() {
+	if (solver_ && this->next != this) {
+		// mark removed propagators by creating "self-loop"
 		solver_->removePost(&messageHandler_);
 		messageHandler_.next = &messageHandler_;
-	}
-	if (!onyMessageHandler && this->next != this) {
 		solver_->removePost(this);
 		this->next = this;
 	}
-		
 }
 
-void ParallelHandler::release() {
-	detach(false);
-	updateModel_ = false;
-	while (!integrated_.empty()) {
-		ClauseHead* c = (ClauseHead*)integrated_.back();
-		integrated_.pop_back();
-		if (c->locked(*solver_) || c->activity().score() > 0) {
-			solver_->addLearnt(c, c->size());
-		}
-		else {
-			c->destroy(solver_, true);
-		}
-	}
+// split-off new guiding path and add it to solve object
+void ParallelHandler::handleSplitMessage() {
+	assert(solver_ && "ParallelHandler::handleSplitMessage(): not attached!");
+	Solver& s = *solver_;
+	s.updateGuidingPath(gp_.path, gp_.pos, gp_.impl);
+	LitVec newGP(gp_.path);
+	s.pushRootLevel();
+	newGP.push_back(~s.decision(s.rootLevel()));
+	++s.stats.parallel->splits;
+	gp_.split = true;
+	ctrl()->pushWork(newGP);
 }
 
-void ParallelHandler::attach() {
-	if (messageHandler_.next == &messageHandler_) {
-		messageHandler_.next = 0;
-		solver_->addPost(&messageHandler_);
-	}
-	if (this->next == this) {
-		this->next = 0;
-		solver_->addPost(this);
-	}
+bool ParallelHandler::handleRestartMessage() {
+	// TODO
+	// we may want to implement some heuristic, like
+	// computing a local var order. 
+	return true;
 }
 
 bool ParallelHandler::simplify(Solver& s, bool sh) {
@@ -714,16 +837,24 @@ bool ParallelHandler::simplify(Solver& s, bool sh) {
 	return false;
 }
 
-bool ParallelHandler::propagateFixpoint(Solver& s) {
+// integrate new information from models
+bool ParallelHandler::MessageHandler::doUpdate(Solver& s, const GP& gp) {
 	// Skip update of model while assumption literal is not yet assigned.
 	// This is necessary during hierarchical optimization because otherwise
 	// integrating a too strong optimum might irrevocably force the assumption literal
 	// which would defeat its purpose.
-	if (updateModel_ && s.isTrue(s.sharedContext()->tagLiteral())) {
-		if (!ctrl()->enumerator()->update(s, disjointPath())) {
+	if (s.isTrue(s.sharedContext()->tagLiteral())) {
+		if (!s.sharedContext()->enumerator()->update(s, gp.split)) {
 			return false;
 		}
-		updateModel_ = false;
+		msg.doUpdate = false;
+	}
+	return true;	
+}
+
+bool ParallelHandler::propagateFixpoint(Solver& s) {
+	if (messageHandler_.pendingUpdate() && !messageHandler_.doUpdate(s, gp_)) {
+		return false;
 	}
 	SharedLiterals* temp[5];
 	uint32 rec;
@@ -736,9 +867,23 @@ bool ParallelHandler::propagateFixpoint(Solver& s) {
 			return false;
 		}
 	} while (rec == 5);
+	if (s.stats.conflicts >= gp_.restart) {
+		ctrl()->requestRestart();
+		gp_.restart *= 2;
+	}
 	return true;
 }
 
+// checks whether s still has a model once all 
+// information from previously found models were integrated 
+bool ParallelHandler::isModel(Solver& s) {
+	assert(s.numFreeVars() == 0);
+	// either no unprocessed updates or still a model after
+	// updates was integrated
+	return !ctrl()->checkUpdates(messageHandler_.msg)
+		|| (messageHandler_.doUpdate(s, gp_) && s.numFreeVars() == 0);
+}
+
 bool ParallelHandler::integrateClauses(Solver& s, SharedLiterals** arr, uint32 num) {
 	assert(!s.hasConflict() && &s == solver_);
 	ClauseCreator::Result ret(0, true);
@@ -773,49 +918,6 @@ void ParallelHandler::add(ClauseHead* h) {
 	}
 }
 
-bool ParallelHandler::isModel(Solver& s) {
-	assert(s.numFreeVars() == 0);
-	if (ctrl()->validModel(s, updateModel_, disjointPath())) {
-		updateModel_ = false;
-		return true;
-	}
-	return false;
-}
-
-bool ParallelHandler::handleSplitMessage(Solver& s) {
-	assert(&s == solver_ && solver_->decisionLevel() > solver_->rootLevel());
-	s.updateGuidingPath(gp_.path, gp_.pos, gp_.impl);
-	LitVec newGP(gp_.path);
-	s.pushRootLevel();
-	newGP.push_back(~s.decision(s.rootLevel()));
-	++s.stats.parallel->splits;
-	setDisjointPath(true);
-	ctrl()->pushWork(newGP);
-	return true;
-}
-
-bool ParallelHandler::handleModelMessage(Solver& s) {
-	assert(&s == solver_); (void)s;
-	updateModel_ = true;
-	return true;
-}
-
-bool ParallelHandler::handleTerminateMessage(Solver& s) {
-	assert(&s == solver_);
-	// handle termination message once and then ignore
-	// all further messages
-	s.stats.parallel->terminated = true;
-	detach(false);
-	s.setStopConflict();
-	return false;
-}
-
-bool ParallelHandler::handleRestartMessage(Clasp::Solver&) {
-	// TODO
-	// we may want to implement some heuristic, like
-	// computing a local var order. 
-	return true;
-}
 /////////////////////////////////////////////////////////////////////////////////////////
 // GlobalQueue
 /////////////////////////////////////////////////////////////////////////////////////////
