@@ -55,44 +55,40 @@ void Preprocessor::updatePreviouslyDefinedAtoms(Var startAtom, bool strong) {
 			a->setLiteral(x);
 		}
 	}
+	LitVec::size_type old = prg_->compute_.size();
 	for (Var i = 1; i < startAtom; ++i) {
 		PrgAtomNode* a   = prg_->atoms_[i];
-		uint32 eq        = prg_->getEqAtom(i);
-		PrgAtomNode* aEq = prg_->atoms_[eq];
-		bool unfreeze    = a->literal().watched();
-		if (unfreeze){ 
+		if (a->literal().watched()) { 
 			Literal x = a->literal(); x.clearWatch(); 
 			a->setLiteral(x); 
+			if (!strong) { a->preds.clear(); }
+			else         { setRootAtom(a->literal(), i); };
 		}
-		if (strong)  {
-			if (eq != i) {
-				a->clearLiteral(true);
-				a->setValue( aEq->value() );
-				a->setLiteral( aEq->literal() );
-			}
-			if (getRootAtom(a->literal()) == varMax) {
-				setRootAtom(a->literal(), eq);
-			}
-			if (!unfreeze) {
+		else if (!a->eq() && a->value() != value_false) {
+			if (strong) {
+				setRootAtom(a->literal(), i);
 				nodes_[i].aSeen = 1;
 				propagateAtomVar((Var)i, a);
-			}			
-		}
-		else if (unfreeze) {
-			a->preds.clear();
-		}
-		else if (aEq->hasVar() && a->value() != value_false) {
-			assert( aEq == a );
-			for (VarVec::size_type b = 0; b != a->posDep.size(); ++b) {
-				PrgBodyNode* body = prg_->bodies_[a->posDep[b]];
-				bool isSupp = body->isSupported();
-				body->onPosPredSupported(Var(i));
-				if (!isSupp && body->isSupported()) {
-					prg_->initialSupp_.push_back(a->posDep[b]);
+			}
+			else {
+				for (VarVec::size_type b = 0; b != a->posDep.size(); ++b) {
+					PrgBodyNode* body = prg_->bodies_[a->posDep[b]];
+					bool isSupp = body->isSupported();
+					body->onPosPredSupported(Var(i));
+					if (!isSupp && body->isSupported()) {
+						prg_->initialSupp_.push_back(a->posDep[b]);
+					}
 				}
 			}
 		}
+		else {
+			assert(!a->eq() || (a->posDep.empty() && a->negDep.empty() && "EQ-Atom in new Program - rule not simplified!"));
+			if (a->posDep.size() + a->negDep.size() + a->preds.size() > 0) {
+				prg_->setCompute(i, false);
+			}
+		}
 	}
+	applyCompute(prg_->compute_, old, strong);
 }
 
 // Back-propagates atoms from the compute statement.
