@@ -324,13 +324,10 @@ bool PrgBodyNode::addPredecessorClauses(SharedContext& ctx, ClauseCreator& gc, c
 // Remove false/true literals
 bool PrgBodyNode::simplifyBody(ProgramBuilder& prg, uint32 bodyId, std::pair<uint32, uint32>& hashes, Preprocessor& pre, bool strong) {
 	hashes.first = hashes.second = 0;
-	bool ok = sumWeights() >= bound();
+	bool ok  = sumWeights() >= bound();
 	uint32 j = 0;
 	Var eq, comp;
-	bool rem;
-	uint32 dirty = 0;
 	for (uint32 i = 0, end = size_; i != end && ok; ++i) {
-		rem = true;
 		Var      a = goals_[i].var();
 		hashes.first += hashId(goals_[i].sign()?-a:a);
 		if ((eq = prg.getEqAtom(a)) != a) {
@@ -341,7 +338,6 @@ bool PrgBodyNode::simplifyBody(ProgramBuilder& prg, uint32 bodyId, std::pair<uin
 			else {
 				a         = comp;
 				goals_[i] = Literal(comp, true);
-				dirty     = posSize_;
 				prg.atoms_[comp]->negDep.push_back(bodyId);
 				VarVec& eqPos = prg.atoms_[eq]->posDep;
 				VarVec::iterator it = std::find(eqPos.begin(), eqPos.end(), bodyId);
@@ -350,7 +346,7 @@ bool PrgBodyNode::simplifyBody(ProgramBuilder& prg, uint32 bodyId, std::pair<uin
 		}
 		Literal p; 
 		ValueRep v = prg.atoms_[a]->value();
-		bool mark  = false;
+		bool mark  = false, rem = true;
 		if (prg.atoms_[a]->hasVar() || strong) {
 			p = goals_[i].sign() ? ~prg.atoms_[a]->literal() : prg.atoms_[a]->literal();
 			v = prg.atoms_[a]->hasVar() ? v : value_false;
@@ -411,31 +407,25 @@ bool PrgBodyNode::simplifyBody(ProgramBuilder& prg, uint32 bodyId, std::pair<uin
 			}
 		}
 	}
-	size_       = j;
-	uint32 posS = 0;
-	for (uint32 i = 0, end = (uint32)size_; i != end; ++i) {
-		prg.vars_.unmark( prg.atoms_[goals_[i].var()]->var() );
-		hashes.second += hashId(goals_[i].sign()?-goals_[i].var():goals_[i].var());
-		posS          += goals_[i].sign() == false;
-	}
-	posSize_ = posS;
-	if (dirty && posS) {
-		// we changed some positive goals to negative ones - 
-		// restore pos | neg partition
-		// find last pos subgoal - must be there otherwise posS == 0
-		j = dirty;
-		while (goals_[--j].sign() == true) { ; }
-		for (uint32 i = 0; ; --j, ++i) {
-			// find first neg subgoal
-			while (i < j && goals_[i].sign() == false) { ++i; }
-			if (i >= j) break;
-			// restore order
-			std::swap(goals_[i], goals_[j]);	
+	// unmark atoms, compute new hash value,
+	// and restore pos | neg partition in case
+	// we changed some positive goals to negative ones 
+	size_    = j;
+	uint32 p = 0, n = j;
+	for (uint32 a, h; p < n;) {
+		if      (!goals_[p].sign())      { h = hashId( (a = goals_[p++].var()));  }
+		else if (goals_[n-1].sign())     { h = hashId(-(a = goals_[--n].var()));  }
+		else /* restore pos|neg order */ {
+			std::swap(goals_[p], goals_[n-1]);
 			if (hasWeights()) {
-				std::swap(extra_.ext->weights[i], extra_.ext->weights[j]);
+				std::swap(extra_.ext->weights[p], extra_.ext->weights[n-1]);
 			}
+			continue;
 		}
+		prg.vars_.unmark( prg.atoms_[a]->var() );
+		hashes.second += h;
 	}
+	posSize_ = p;
 	assert(sumWeights() >= bound() || !ok);
 	if (!ok) {          // body is false...
 		setIgnore(true);  // ...and therefore can be ignored
@@ -649,17 +639,16 @@ bool PrgBodyNode::propagateFalse(uint32 myId, AtomList& prgAtoms) {
 /////////////////////////////////////////////////////////////////////////////////////////
 class ProgramBuilder::CycleChecker {
 public:
-	CycleChecker(const AtomList& prgAtoms, const BodyList& prgBodies, bool check)
+	CycleChecker(const AtomList& prgAtoms, const BodyList& prgBodies, bool check, uint32 startScc)
 		: atoms_(prgAtoms)
 		, bodies_(prgBodies)
 		, count_(0)
-		, sccs_(0)
+		, sccs_(startScc)
 		, check_(check) {
 	}
 	void visit(PrgBodyNode* b) { if (check_) visitDfs(b, true); }
 	void visit(PrgAtomNode* a) { if (check_) visitDfs(a, false); }
-	bool   hasCycles() const     { return sccs_ > 0; }
-	uint32 sccs()      const     { return sccs_; }
+	uint32 sccs()              const { return sccs_; }
 	const AtomList& sccAtoms() const { return sccAtoms_; }
 private:
 	CycleChecker& operator=(const CycleChecker&);
@@ -815,10 +804,7 @@ void PreproStats::accu(const PreproStats& o) {
 /////////////////////////////////////////////////////////////////////////////////////////
 ProgramBuilder::ProgramBuilder() : minimize_(0), incData_(0), graph_(0) { }
 ProgramBuilder::~ProgramBuilder() { disposeProgram(true); }
-ProgramBuilder::Incremental::Incremental() : startAtom_(1), startVar_(1), startAux_(1) {}
-bool ProgramBuilder::Incremental::frozen(Var atom) const {
-	return std::find(freeze_.begin(), freeze_.end(), atom) != freeze_.end();
-}
+ProgramBuilder::Incremental::Incremental() : startAtom_(1), startVar_(1), startAux_(1), startScc_(0) {}
 
 void ProgramBuilder::disposeProgram(bool force) {
 	std::for_each( bodies_.begin(), bodies_.end(), DestroyObject() );
@@ -916,9 +902,14 @@ ProgramBuilder& ProgramBuilder::updateProgram() {
 	// clean up atoms
 	assert(incData_->startAux_ <= atoms_.size());
 	for (VarVec::size_type i = 0; i != incData_->startAux_; ++i) {
-		if (atoms_[i]->eq() && atoms_[i]->eqNode() >= incData_->startAux_) {
+		if (atoms_[i]->eq() && getEqAtom(i) >= incData_->startAux_) {
 			// atom i is equivalent to some aux atom 
-			std::swap(atoms_[i], atoms_[atoms_[i]->eqNode()]);
+			// make i the new root
+			PrgAtomNode* eq = atoms_[getEqAtom(i)];
+			assert(!eq->eq());
+			eq->posDep.clear(); eq->negDep.clear(); eq->preds.clear();
+			atoms_[i]->assign(eq);
+			eq->setEq(i);
 		}
 		VarVec().swap(atoms_[i]->posDep);
 		VarVec().swap(atoms_[i]->negDep);
@@ -942,6 +933,9 @@ ProgramBuilder& ProgramBuilder::updateProgram() {
 
 ProgramBuilder& ProgramBuilder::addRule(PrgRule& r) {
 	// simplify rule, mark literals
+	if (startAtom() > 1) {
+		updateRule(r);
+	}
 	PrgRule::RData rd = r.simplify(ruleState_);
 	if (r.type() != ENDRULE) {     // rule is relevant
 		upRules(r.type(), 1);
@@ -970,6 +964,15 @@ ProgramBuilder& ProgramBuilder::addRule(PrgRule& r) {
 	return *this;
 }
 
+void ProgramBuilder::updateRule(PrgRule& r) {
+	for (WeightLitVec::iterator it = r.body.begin(), end = r.body.end(); it != end; ++it) {
+		Var id = it->first.var();
+		if (id < atoms_.size() && getEqAtom(id) != id) {
+			it->first = Literal(getEqAtom(id), it->first.sign());
+		}
+	}
+}
+
 bool ProgramBuilder::handleNatively(const PrgRule& r, const PrgRule::RData& rd) const {
 	ExtendedRuleMode m = eqOpts_.erMode;
 	if (m == mode_native || m == mode_transform_integ || r.type() == BASICRULE || r.type() == OPTIMIZERULE) {
@@ -1007,7 +1010,7 @@ void ProgramBuilder::addRuleImpl(PrgRule& r, const PrgRule::RData& rd) {
 		for (VarVec::const_iterator it = r.heads.begin(), end = r.heads.end(); it != end; ++it) {
 			if (b.first->value() != value_false) {
 				PrgAtomNode* a = resize(*it);
-				assert( (!a->hasVar() || (incData_ && incData_->frozen(*it))) && "Cannot extend definition of atom!");
+				assert( (!a->hasVar() || a->frozen()) && "Cannot extend definition of atom!");
 				if (r.body.empty()) a->setIgnore(true);
 				// Note: b->heads may now contain duplicates. They are removed in PrgBodyNode::buildHeadSet.
 				HeadType t = b.first->addHead(*it, r.type());
@@ -1268,12 +1271,14 @@ bool ProgramBuilder::transformProgram(bool checkSccs) {
 	if (checkSccs && graph_.get() == 0) {
 		graph_ = new SharedDependencyGraph();
 	}
-	CycleChecker c(atoms_, bodies_, checkSccs);
+	uint32 scc = 0;
 	if (incData_) {
 		for (VarVec::const_iterator it = incData_->unfreeze_.begin(), end = incData_->unfreeze_.end(); it != end; ++it) {
 			atoms_.push_back(atoms_[*it]);
 		}
+		scc = incData_->startScc_;
 	}
+	CycleChecker c(atoms_, bodies_, checkSccs, scc);
 	ctx_->startAddConstraints();
 	SatPreprocessor* p = 0;
 	if (ctx_->satPrepro.get() && ctx_->satPrepro->limit(std::max(stats.atoms, stats.bodies))) {
@@ -1281,13 +1286,16 @@ bool ProgramBuilder::transformProgram(bool checkSccs) {
 	}
 	bool ret = addConstraints(c);
 	if (p) ctx_->satPrepro.reset(p);
-	if (checkSccs) stats.sccs = c.sccs();
-	if (ret && c.hasCycles()) {
+	if (checkSccs) stats.sccs = (c.sccs()-scc);
+	if (ret && !c.sccAtoms().empty()) {
 		uint32 oldNodes = graph_->nodes();
 		graph_->addSccs(*ctx_, c.sccAtoms(), atoms_, bodies_);
 		stats.ufsNodes  = graph_->nodes()-oldNodes;
 	}
-	if (incData_) atoms_.resize( atoms_.size() - incData_->unfreeze_.size() );
+	if (incData_) {
+		atoms_.resize( atoms_.size() - incData_->unfreeze_.size() );
+		incData_->startScc_ = c.sccs();
+	}
 	return ret;
 }
 
