@@ -27,7 +27,7 @@
 #include <cassert>
 #include <stdio.h>
 #include <stdlib.h>
-#ifdef _MSC_VER
+#ifdef _WIN32
 #include <io.h>
 #include <fcntl.h>
 #include <sys/stat.h>
@@ -131,6 +131,26 @@ bool StreamSource::readLine( char* buf, unsigned maxSize ) {
 	*buf = 0;
 	return maxSize != 0;
 }
+
+bool readLine(StreamSource& in, PodVector<char>::type& buf ) {
+	char   buffer[1024];
+	uint32 i;
+	buf.clear();
+	for (i = 0;;) {
+		buffer[i] = *in;
+		++in;
+		if (buffer[i] == '\n' || buffer[i] == '\0') {
+			buf.insert(buf.end(), buffer, buffer+i+1);
+			buf.back() = '\0';
+			break;
+		}
+		if (++i == 1024) {
+			buf.insert(buf.end(), buffer, buffer+i);
+			i = 0;
+		}
+	}
+	return buffer[i] == '\n';
+}
 	
 namespace {
 /////////////////////////////////////////////////////////////////////////////////////////
@@ -269,16 +289,17 @@ bool LparseReader::readBody(uint32 lits, uint32 neg, bool readWeights) {
 
 bool LparseReader::readSymbolTable() {
 	int a = -1;
+	PodVector<char>::type buf;
+	buf.reserve(1024);
 	while (source_->parseInt(a) && a != 0) {
 		if (a < 1) {
 			throw ReadError(source_->line(), "Symbol Table: Atom id out of bounds!");
 		}
-		char buffer[1024];
 		source_->skipWhite();
-		if (!source_->readLine(buffer, 1024)) {
+		if (!readLine(*source_, buf)) {
 			throw ReadError(source_->line(), "Symbol Table: Atom name too long or end of file!");
 		}
-		api_->setAtomName(a, buffer);
+		api_->setAtomName(a, &buf[0]);
 		skipAllWhite(*source_);
 	}
 	if (a != 0) {
@@ -409,7 +430,7 @@ bool parseDimacsImpl(std::istream& prg, SharedContext& ctx, bool assertPure, Obj
 					nc.start();
 					if (!sat && top != w) {
 						// soft clause
-						if (currentClause.size() > 1 || true) {
+						if (currentClause.size() > 1) {
 							flags.push_back(0);
 							currentClause.push_back(posLit(++lastVar));
 							opt.lits.push_back(WeightLiteral(currentClause.back(), w));
@@ -450,7 +471,7 @@ bool parseDimacsImpl(std::istream& prg, SharedContext& ctx, bool assertPure, Obj
 			}
 		}
 		else if (match(in, "p ", false)) {
-			if (match(in, "cnf", false) || (wcnf=match(in, "wcnf", false))) {
+			if (match(in, "cnf", false) || (wcnf=match(in, "wcnf", false)) == true) {
 				if (!in.parseInt(numVars) || !in.parseInt(numClauses) ) {
 					ERROR("Bad parameters in the problem line!");
 				}
@@ -836,11 +857,19 @@ bool OPBParser::finalize() {
 void OPBParser::addConstraint(const char* relOp, int rhs, uint32 var) {
 	if (good()) {
 		if (relOp != 0) {
-			// var == lhs >= rhs
-			WeightConstraint::newWeightConstraint(ctx_, posLit(var), lhs_, rhs);
-			if (relOp[0] == '=') {
-				// ~var == lhs >= rhs+1
-				WeightConstraint::newWeightConstraint(ctx_, negLit(var), lhs_, rhs+1);
+			if (relOp[0] == '>') {
+				assert(relOp[1] == '=');
+				// var == [lhs >= rhs]
+				WeightConstraint::newWeightConstraint(ctx_, posLit(var), lhs_, rhs);	
+			}
+			else {
+				assert(relOp[0] == '=');
+				// replace var == [lhs = rhs] with
+				// var  == [lhs >= rhs]   and
+				// ~var == [lhs >= rhs+1]
+				WeightLitVec temp(lhs_); // copy because newWeightConstraint() removes assigned literals 
+				WeightConstraint::newWeightConstraint(ctx_, posLit(var), lhs_, rhs) &&
+				WeightConstraint::newWeightConstraint(ctx_, negLit(var), temp, rhs+1);
 			}
 		}
 		else {
