@@ -26,6 +26,9 @@ namespace Clasp {
 /////////////////////////////////////////////////////////////////////////////////////////
 // ShortImplicationsGraph::ImplicationList
 /////////////////////////////////////////////////////////////////////////////////////////
+namespace {
+	inline bool isBinary(const Literal* imp) { return imp->watched(); }
+}
 #ifndef DISABLE_MULTI_THREADING
 struct ShortImplicationsGraph::Block {
 	typedef std::atomic<uint32> atomic_size;
@@ -117,11 +120,11 @@ void ShortImplicationsGraph::ImplicationList::addLearnt(Literal p, Literal q) {
 bool ShortImplicationsGraph::ImplicationList::hasLearnt(Literal q, Literal r) const {
 	const bool binary = isSentinel(r);
 	FOR_EACH_LEARNT(*this, imp) {
-		if (imp->watched()) {
-			if (imp[0] == q || imp[1] == r) { return true; }
-		}
-		else if (!binary && ((imp[0] == q && imp[1] == r) || (imp[1] == q && imp[0] == r))) {
-			return true;
+		if (imp[0] == q || imp[0] == r) {
+			// binary clause subsumes new bin/tern clause
+			if (isBinary(imp))                           { return true; }
+			// existing ternary clause subsumes new tern clause
+			if (!binary && (imp[1] == q || imp[1] == r)) { return true; }
 		}
 	}
 	return false;
@@ -264,8 +267,8 @@ bool ShortImplicationsGraph::propagate(Solver& s, Literal p) const {
 	}
 	// combined short BCP on learnt clauses
 	FOR_EACH_LEARNT(x, imp) {
-		if (!s.isTrue(q = imp[0]) && (imp->watched() || !s.isTrue(r = imp[1]))) {
-			if (!imp->watched()) {
+		if (!s.isTrue(q = imp[0]) && (isBinary(imp) || !s.isTrue(r = imp[1]))) {
+			if (!isBinary(imp)) {
 				if (s.isFalse(q)) { q = r; r = imp[0]; }
 				if (s.isFalse(r) && !s.force(q, Antecedent(p, ~r))) {
 					return false;
@@ -295,8 +298,8 @@ bool ShortImplicationsGraph::reverseArc(const Solver& s, Literal p, uint32 maxLe
 		}
 	}
 	FOR_EACH_LEARNT(x, imp) {
-		if (isRevLit(s, imp[0], maxLev) && (imp->watched() || isRevLit(s, imp[1], maxLev))) {
-			out = !imp->watched() ? Antecedent(~imp[0], ~imp[1]) : Antecedent(~imp[0]);
+		if (isRevLit(s, imp[0], maxLev) && (isBinary(imp) || isRevLit(s, imp[1], maxLev))) {
+			out = !isBinary(imp) ? Antecedent(~imp[0], ~imp[1]) : Antecedent(~imp[0]);
 			return true;
 		}
 	}
