@@ -206,6 +206,7 @@ SolverStrategies::SolverStrategies()
 ////////////////////////////////////////////////////////////////////////////////////////
 Solver::Solver() 
 	: strategy_()
+	, ccMin_(0)
 	, smallAlloc_(new SmallClauseAlloc)
 	, shared_(0)
 	, randHeuristic_(0)
@@ -218,12 +219,11 @@ Solver::Solver()
 	, rootLevel_(0)
 	, btLevel_(0)
 	, tagged_(0)
+	, lbdTime_(0)
 	, shuffle_(false) {
 	Var sentVar = assign_.addVar();
 	assign_.setValue(sentVar, value_true);
 	markSeen(sentVar);
-	ccMin_ = 0;
-	lbdTime_ = 0;
 }
 
 Solver::~Solver() {
@@ -237,8 +237,6 @@ void Solver::freeMem() {
 	learnts_.clear();
 	setEnumerationConstraint(0);
 	PodVector<WatchList>::destruct(watches_);
-	delete levConflicts_;
-	delete randHeuristic_;
 	// free undo lists
 	// first those still in use
 	for (DecisionLevels::size_type i = 0; i != levels_.size(); ++i) {
@@ -252,6 +250,12 @@ void Solver::freeMem() {
 	}
 	delete smallAlloc_;
 	delete ccMin_;
+	delete levConflicts_;
+	delete randHeuristic_;
+	smallAlloc_   = 0;
+	ccMin_        = 0;
+	levConflicts_ = 0;
+	randHeuristic_= 0;
 }
 
 SatPreprocessor* Solver::satPrepro() const { return shared_->satPrepro.get(); }
@@ -271,6 +275,9 @@ void Solver::startInit(SharedContext* ctx, uint32 numConsGuess) {
 	assign_.trail.reserve(numVars());
 	constraints_.reserve(numConsGuess/2);
 	levels_.reserve(25);
+	if (smallAlloc_ == 0) {
+		smallAlloc_ = new SmallClauseAlloc();
+	}
 	if (undoHead_ == 0) {
 		for (uint32 i = 0; i != 25; ++i) { 
 			undoFree(new ConstraintDB(10)); 
@@ -618,11 +625,14 @@ void Solver::setConflict(Literal p, const Antecedent& a, uint32 data) {
 }
 
 bool Solver::assume(const Literal& p) {
-	assert( value(p.var()) == value_free && decisionLevel() != assign_.maxLevel());
-	++stats.choices;
-	levels_.push_back(DLevel(numAssignedVars(), 0));
-	if (levConflicts_) levConflicts_->push_back( (uint32)stats.conflicts );
-	return force(p, Antecedent());  // always true
+	if (value(p.var()) == value_free) {
+		assert(decisionLevel() != assign_.maxLevel());
+		++stats.choices;
+		levels_.push_back(DLevel(numAssignedVars(), 0));
+		if (levConflicts_) levConflicts_->push_back( (uint32)stats.conflicts );
+		return assign_.assign(p, decisionLevel(), Antecedent());
+	}
+	return isTrue(p);
 }
 
 bool Solver::propagate() {
@@ -963,7 +973,7 @@ void Solver::otfs(Antecedent& lhs, const Antecedent& rhs, Literal p, bool final)
 
 ClauseHead* Solver::otfsRemove(ClauseHead* c, const LitVec* newC) {
 	ConstraintType t = c->type();
-	bool remStatic   = !newC || newC->size() <= 3;
+	bool remStatic   = !newC || (newC->size() <= 3 && shared_->unique());
 	if ((t == Constraint_t::static_constraint && remStatic) || 
 	    (t == Constraint_t::learnt_conflict || t == Constraint_t::learnt_loop)) {
 		ConstraintDB& db = (t == Constraint_t::static_constraint ? constraints_ : learnts_);
@@ -1002,16 +1012,21 @@ uint32 Solver::finalizeConflictClause(ClauseHead* rhs) {
 	rhs             = rhs && cc_.size() <= conflict_.size() ? rhs : 0;
 	if (rhs) {
 		markSeen(cc_[0].var());
-		uint32 m = static_cast<uint32>((conflict_.size() - cc_.size()) + 1);
-		for (LitVec::size_type i = 0; i != conflict_.size() && m; ++i) {
-			m -= !seen(conflict_[i].var());
-		}
-		if ( (rhs = m ? otfsRemove(rhs, &cc_) : 0) != 0 ) {
+		// check if rhs is subsumed by cc_
+		// NOTE: at this point rhs == ~conflict_
+		uint32 lhsOpen = (uint32)cc_.size();
+		for (LitVec::size_type i = 0; i != conflict_.size() && lhsOpen; ++i) {
+			// NOTE: at this point the DB might not be fully simplified,
+			//       e.g. because of mt or lookahead, hence we must explicitly
+			//       check for literals assigned on DL 0
+			lhsOpen -= level(conflict_[i].var()) > 0 && seen(conflict_[i].var());
+  	}
+		if ( (rhs = lhsOpen == 0 ? otfsRemove(rhs, &cc_) : 0) != 0 ) {
 			// rhs is subsumed by cc_ but could not be deleted -
 			// try to strengthen.
 			ClauseHead::BoolPair x(true, false);
 			for (LitVec::size_type i = 0; i != conflict_.size() && x.first; ++i) {
-				if (!seen(conflict_[i].var())) {
+				if (!seen(conflict_[i].var()) || level(conflict_[i].var()) == 0) {
 					x = rhs->strengthen(*this, ~conflict_[i], false);
 				}
 			}
@@ -1396,8 +1411,8 @@ ValueRep Solver::search(RestartLimit& rlimit, LearntLimit& dlimit, double randPr
 			}
 		} while (decideNextBranch());
 		// found a model candidate
-		assert(numFreeVars() == 0);
-	} while (!post_.isModel(*this));
+		assert(numFreeVars() == 0 || hasConflict());
+	} while (hasConflict() || !post_.isModel(*this));
 	stats.addModel(decisionLevel());
 	if (satPrepro()) {
 		satPrepro()->extendModel(assign_, unconstr_);
