@@ -110,6 +110,7 @@ class ProgramBuilderTest : public CppUnit::TestFixture {
 	CPPUNIT_TEST(testWeightlessAtomsInWeightRule);
 	CPPUNIT_TEST(testSimplifyToNormal);
 	CPPUNIT_TEST(testSimplifyToCardBug);
+	CPPUNIT_TEST(testSimplifyCompBug);
 
 	CPPUNIT_TEST(testBPWeight);
 
@@ -1121,6 +1122,32 @@ public:
 		CPPUNIT_ASSERT(ctx.master()->numFreeVars() == 0);
 	}
 
+	void testSimplifyCompBug() {
+		// x1 :- not x2.
+		// x1 :- x2.
+		// x2 :- not x3.
+		// x3 :- not x1.
+		// a. b. f.
+		// x4 :- a, b, x2, e, f.
+		builder.startProgram(ctx, ProgramBuilder::EqOptions().iterations(1))
+			.setAtomName(1, "x1").setAtomName(2, "x2").setAtomName(3, "x3")
+			.setAtomName(4, "a").setAtomName(5, "b").setAtomName(6, "e").setAtomName(7, "f")
+			.startRule().addHead(1).addToBody(2, false).endRule()
+			.startRule().addHead(1).addToBody(2, true).endRule()
+			.startRule().addHead(2).addToBody(3, false).endRule()
+			.startRule().addHead(3).addToBody(1, false).endRule()
+			.startRule().addHead(4).addHead(5).addHead(7).endRule()
+			.startRule(CHOICERULE).addHead(6).endRule()
+			.startRule().addHead(8).addToBody(4, true).addToBody(5, true).addToBody(2, true).addToBody(6, true).addToBody(7, true).endRule()
+		;
+		CPPUNIT_ASSERT_EQUAL(true, builder.endProgram());
+		PrgAtomNode* x = builder.getAtom(8);
+		PrgBodyNode* B = builder.getBody(x->preds[0].node());
+		CPPUNIT_ASSERT(B->size() == 2 && B->posSize() == 1 && B->negSize() == 1);
+		CPPUNIT_ASSERT(B->pos(0) == 6);
+		CPPUNIT_ASSERT(B->neg(0) == 3);
+	}
+
 	void testBPWeight() {
 		// {a, b, c, d}.
 		// x :-  2 [a=1, b=2, not c=2, d=1].
